"""Track is the common ancestor for all specialized tracks of Specialized Points.

To create a new specialized track, you must do the following:
- Create your specialized XYZSpot class:
    - Make it inherit from Spot
    - implement (static function) Spot.get_extra_features_name()
    - implement (method) Spot.get_extra_features()
    (see their doc for additional help)
    
- Create your specialized XYZTrack class:
    - Make it inherit from Track
    - implement (static function) Track.track_df_to_track_list()

- In gen_track.py:
    - import your new specialized track,
    - import your new specialized spot,
    - add them to the SPOT_AND_TRACK_MAPPING dict,
"""

from __future__ import annotations
from typing import TypeVar, Generic
from abc import abstractmethod, ABC

import pandas as pd
from .spot import Spot
from .cell_spot import CellSpot
from .mid_body_spot import MidBodySpot

T = TypeVar("T", MidBodySpot, CellSpot)


class Track(ABC, Generic[T]):
    """
    Class used for both cell and mid-body tracks.

    Parameters
    ----------
    track_id : int
        Track id.
    """

    def __init__(self, track_id: int) -> None:
        self.track_id = track_id

        self.spots: dict[int, T] = {}
        self.length = 0  # = len(self.spots)

    def add_spot(self, spot: Spot) -> None:
        """
        Add spot to track.
        """
        self.spots[spot.frame] = spot
        spot.track_id = self.track_id
        self.length += 1
        # Add children recursively
        if hasattr(spot, "child_spot") and spot.child_spot is not None:
            self.add_spot(spot.child_spot)

    @staticmethod
    @abstractmethod
    def track_df_to_track_list(
        track_df: pd.DataFrame,
        spots: dict[int, list[Spot]],
    ) -> list[Track]:
        """Implemented in children classes."""
        return []



import numpy as np

from cnn_framework.utils.data_sets.dataset_output import DatasetOutput
from cnn_framework.utils.preprocessing import normalize_array

from ..cnn_data_set import CnnDataSet


class MetaphaseCnnDataSet(CnnDataSet):
    """
    Only difference with parent class is normalization which is applied for
    each image independently.
    This is not something to be done, but current old model was developed that way.
    """

    def generate_images(self, filename: str) -> DatasetOutput:
        """Generate images from the given filename."""
        idx = int(filename.split(".")[0])
        nucleus_image = normalize_array(self.data[idx], None)  # CYX
        nucleus_image = np.moveaxis(nucleus_image, 0, -1)  # YXC
        # Define any target value
        target_array = np.zeros(self.params.nb_classes)
        target_array[0] = 1
        # Construct output
        return DatasetOutput(
            input=nucleus_image,
            target_array=target_array,
        )



import os
import pickle
from typing import Optional
import numpy as np
from aicsimageio.writers import OmeTiffWriter
from tqdm import tqdm

from ..factories.mid_body_detection_factory import MidBodyDetectionFactory

from ..utils.mitosis_track import MitosisTrack
from ..utils.cell_track import CellTrack


def perform_mid_body_detection(
    raw_video: np.ndarray,
    video_name: str,
    exported_mitoses_dir: str,
    exported_tracks_dir: str,
    movies_save_dir: Optional[str] = None,
    save: bool = True,
    parallel_detection: bool = False,
    detection_method: str = "difference_gaussian",
    target_mitosis_id: Optional[int] = None,
) -> list[MitosisTrack]:
    """Perform mid-body detection on mitosis tracks.

    Parameters
    ----------
    raw_video : np.ndarray
        Raw video to extract mitosis movies from.
    video_name : str
        Name of the video.
    exported_mitoses_dir : str
        Directory where mitosis tracks are saved.
    exported_tracks_dir : str
        Directory where cell tracks are saved.
    movies_save_dir : Optional[str], optional
        Directory where mitosis movies are saved, by default None.
    save : bool, optional
        Save updated mitosis tracks, by default True.
    parallel_detection : bool, optional
        Perform detection in parallel, by default False.
    detection_method : str, optional
        Detection method to use, by default "difference_gaussian".
    target_mitosis_id : Optional[int], optional
        Target mitosis id to perform mid-body detection on, by default None.

    Returns
    -------
    list[MitosisTrack]
        List of updated mitosis tracks.
    """
    mitosis_tracks: list[MitosisTrack] = []
    # Iterate over "bin" files in exported_mitoses_dir
    for state_path in os.listdir(exported_mitoses_dir):
        # Ignore if not for current video
        if video_name not in state_path:
            continue
        # Load mitosis track
        with open(os.path.join(exported_mitoses_dir, state_path), "rb") as f:
            mitosis_track = MitosisTrack.load(f)

        # Add mitosis track to list
        mitosis_tracks.append(mitosis_track)

    # Load cell tracks
    cell_tracks: list[CellTrack] = []
    # Iterate over "bin" files in exported_tracks_dir
    video_exported_tracks_dir = os.path.join(exported_tracks_dir, video_name)
    for state_path in os.listdir(video_exported_tracks_dir):
        # Load mitosis track
        with open(
            os.path.join(video_exported_tracks_dir, state_path), "rb"
        ) as f:
            cell_track = CellTrack.load(f)
            cell_tracks.append(cell_track)

    print("\n### MID-BODY DETECTION ###")

    # Generate movie for each mitosis and save
    print("Performing mid-body detection.")
    mid_body_detector = MidBodyDetectionFactory()
    for i, mitosis_track in enumerate(tqdm(mitosis_tracks)):

        if (
            isinstance(target_mitosis_id, int)
            and mitosis_track.id != target_mitosis_id
        ):
            print(
                f"\nTrack {i+1}/{len(mitosis_tracks)}, Mitosis id {mitosis_track.id} - Skipped"
            )
            continue

        # Generate mitosis movie
        mitosis_movie, mask_movie = mitosis_track.generate_video_movie(
            raw_video
        )  # TYXC, TYX

        # Search for mid-body in mitosis movie
        mid_body_detector.update_mid_body_spots(
            mitosis_track,
            mitosis_movie,
            cell_tracks,
            parallel_detection=parallel_detection,
            detection_method=detection_method,
        )

        # Save updated mitosis track
        if save:
            daughter_track_ids = ",".join(
                [str(d) for d in mitosis_track.daughter_track_ids]
            )
            state_path = f"{video_name}_mitosis_{mitosis_track.id}_{mitosis_track.mother_track_id}_to_{daughter_track_ids}.bin"
            save_path = os.path.join(
                exported_mitoses_dir,
                state_path,
            )
            with open(save_path, "wb") as f:
                pickle.dump(mitosis_track, f)

        if movies_save_dir:
            # Save mitosis movie
            final_mitosis_movie = mitosis_track.add_mid_body_movie(
                mitosis_movie, mask_movie
            )  # TYX C=C+1
            image_save_path = os.path.join(
                movies_save_dir,
                f"{video_name}_mitosis_{mitosis_track.id}_{mitosis_track.mother_track_id}_to_{daughter_track_ids}.tiff",
            )
            # Transpose to match TCYX
            final_mitosis_movie = np.transpose(
                final_mitosis_movie, (0, 3, 1, 2)
            )
            OmeTiffWriter.save(
                final_mitosis_movie, image_save_path, dim_order="TCYX"
            )

    return mitosis_tracks



"""Playground to run cell segmentation."""

import os
from typing import Optional

from cnn_framework.utils.readers.tiff_reader import TiffReader

from cut_detector._widget import video_whole_process
from cut_detector.data.tools import get_data_path


def main(
    image_path: Optional[str] = get_data_path("videos"),
):
    """
    Parameters
    ----------
    image_path : str
        Path to the image to process.
    """
    # If image_path or model_path are directories, take their first file
    if os.path.isdir(image_path):
        image_path = os.path.join(image_path, os.listdir(image_path)[0])

    # Read image and preprocess if needed
    image = TiffReader(image_path, respect_initial_type=True).image  # TCZYX
    image = image.squeeze()  # CTYX
    image = image.transpose(1, 2, 3, 0)  # TYXC

    image_name = os.path.basename(image_path).split(".")[0]

    spots_dir_name = os.path.join(get_data_path("spots"), image_name)
    tracks_dir_name = os.path.join(get_data_path("tracks"), image_name)

    video_whole_process(
        image,
        image_name,
        default_model_check_box=True,
        segmentation_model="",
        save_check_box=False,
        movies_save_dir="",
        spots_dir_name=spots_dir_name,
        tracks_dir_name=tracks_dir_name,
        mitoses_dir_name=get_data_path("mitoses"),
    )


if __name__ == "__main__":
    main()



import numpy as np

from cut_detector.utils.tools import apply_hmm


def get_category_from_name(initial_category, mode):
    """
    From 10 initial categories to 3 categories: no cut, cut, two cuts.
    Mode is either microtubules or membrane.
    """
    initial_category = int(initial_category)
    initial_category = initial_category % 5

    if mode == "microtubules":
        if initial_category in [0]:
            category = 0
        elif initial_category in [1, 3]:
            category = 1
        elif initial_category in [2, 4]:
            category = 2

    elif mode == "membrane":
        if initial_category in [0, 1, 2]:
            category = 0
        elif initial_category in [3, 4]:
            category = 1

    else:
        raise ValueError(f"Unknown mode: {mode}")

    return category


def get_first_index(lst, value):
    """
    Get index of first occurrence of given value in list, or any value higher.
    """
    if max(lst) < value:
        return len(lst)
    for i, v in enumerate(lst):
        if v >= value:
            return i
    raise ValueError("Value not found in list")


def evaluate_frame_error(
    predictions,
    names,
    hmm_bridges_parameters_file,
    target_category=1,
):
    # Store categories in a dictionary
    classification = {}
    for name, prediction in zip(names, predictions):
        category = name.split("_c")[1][0]
        core_name, frame = name.split("_c")[0].rsplit("_", 1)
        if core_name not in classification:
            classification[core_name] = {"predicted": {}, "ground_truth": {}}
        classification[core_name]["ground_truth"][int(frame)] = (
            get_category_from_name(category, mode="microtubules")
        )
        classification[core_name]["predicted"][int(frame)] = np.argmax(
            prediction
        )

    # Load HMM model to smooth predictions
    hmm_parameters = np.load(hmm_bridges_parameters_file)

    frame_differences = []
    for mitosis_name, mitosis_classification in classification.items():
        frames = list(mitosis_classification["ground_truth"].keys())
        frames.sort()
        ground_truth = [
            mitosis_classification["ground_truth"][frame] for frame in frames
        ]
        predicted = [
            mitosis_classification["predicted"][frame] for frame in frames
        ]
        corrected_predicted = apply_hmm(hmm_parameters, predicted)
        # Get index of first occurrence of target_category
        target_category_index = get_first_index(ground_truth, target_category)
        predicted_target_category_index = get_first_index(
            corrected_predicted, target_category
        )
        frame_difference = (
            predicted_target_category_index - target_category_index
        )
        if abs(frame_difference) > 5:
            print(f"Frame difference: {frame_difference} on {mitosis_name}")
        frame_differences.append(frame_difference)

    print(
        f"Mean frame difference: {np.mean(frame_differences)} +/- {np.std(frame_differences)} on {len(frame_differences)} mitoses."
    )



import torch

from cnn_framework.utils.data_loader_generators.classifier_data_loader_generator import (
    ClassifierDataLoaderGenerator,
)
from cnn_framework.utils.model_managers.cnn_model_manager import (
    CnnModelManager,
)
from cnn_framework.utils.data_managers.default_data_manager import (
    DefaultDataManager,
)
from cnn_framework.utils.metrics.classification_accuracy import (
    ClassificationAccuracy,
)
from cnn_framework.utils.parsers.cnn_parser import CnnParser

from cut_detector.utils.mitosis_track_generation.metaphase_cnn_model import (
    MetaphaseCnnModel,
)
from cut_detector.utils.mitosis_track_generation.metaphase_cnn_model_params import (
    MetaphaseCnnModelParams,
)

from developers.metaphase_cnn.data_set import MetaphaseCnnDataSet


def main(params):
    loader_generator = ClassifierDataLoaderGenerator(
        params, MetaphaseCnnDataSet, DefaultDataManager
    )
    _, _, test_dl = loader_generator.generate_data_loader()

    # Model definition
    # Load pretrained model
    model = MetaphaseCnnModel(
        nb_classes=params.nb_classes,
        nb_input_channels=len(params.c_indexes) * len(params.z_indexes),
        encoder_name=params.encoder_name,
    )
    model.load_state_dict(torch.load(params.model_load_path))

    manager = CnnModelManager(model, params, ClassificationAccuracy)

    manager.predict(test_dl)

    manager.write_useful_information()


if __name__ == "__main__":
    parser = CnnParser()
    args = parser.arguments_parser.parse_args()

    parameters = MetaphaseCnnModelParams()
    parameters.update(args)

    main(parameters)



import os
import xmltodict

from cut_detector.constants.annotations import NAMES_DICTIONARY
from cut_detector.utils.tools import get_video_path


def check_file(xml_folder: str, mitoses_path: str) -> None:
    """Check the xml files in the given folder.

    Parameters
    ----------
    xml_folder : str
        Path to the folder containing xml CellCounter files.
    mitoses_path : str
        Path to the mitoses folder.
    """
    for xml_name in os.listdir(xml_folder):
        print(f"Processing {xml_name}...")
        xml_path = os.path.join(xml_folder, xml_name)

        # Read useful information from xml file
        with open(xml_path) as fd:
            doc = xmltodict.parse(fd.read())

        # Try to find video
        video_name = doc["CellCounter_Marker_File"]["Image_Properties"][
            "Image_Filename"
        ]
        video_path = get_video_path(video_name, mitoses_path)
        if video_path is None:
            raise FileNotFoundError(f"Video {video_name} not found.")

        # Check ids
        for i, type_data in enumerate(
            doc["CellCounter_Marker_File"]["Marker_Data"]["Marker_Type"]
        ):
            assert i == int(type_data["Type"]) - 1  # order should be kept
            if (
                "Name" not in type_data
                or type_data["Name"] not in NAMES_DICTIONARY
            ):
                old_name = type_data["Name"] if "Name" in type_data else ""
                new_name = list(NAMES_DICTIONARY.keys())[
                    list(NAMES_DICTIONARY.values()).index(i)
                ]
                print(
                    f"Class '{old_name}' not found, modified to '{new_name}'."
                )
                # Update accordingly
                doc["CellCounter_Marker_File"]["Marker_Data"]["Marker_Type"][
                    i
                ]["Name"] = new_name
            else:
                assert NAMES_DICTIONARY[type_data.Name] == i

        with open(xml_path, "w") as result_file:
            result_file.write(xmltodict.unparse(doc, pretty=True))


if __name__ == "__main__":
    XML_FOLDER = ""
    MITOSES_PATH = ""

    check_file(XML_FOLDER, MITOSES_PATH)



import glob
import os
from pathlib import Path
import numpy as np

from cnn_framework.utils.readers.tiff_reader import TiffReader
from cnn_framework.utils.display_tools import display_progress
from cnn_framework.utils.enum import ProjectMethods
from cnn_framework.utils.readers.utils.projection import Projection
from cnn_framework.utils.tools import save_tiff


def extract_frames(
    save_dir: str,
    data_dir: str,
    time_dimension: str,
    frames_per_file: str,
    projection: list[Projection],
) -> None:
    """Extract random frames from tiff files in a folder.

    Parameters
    ----------
    save_dir : str
        Path to the folder where to save the images.
    data_dir : str
        Path to the folder containing the tiff files.
    time_dimension : int
        Time dimension in the tiff file.
    frames_per_file : int
        How many frames to extract from each video.
    projection : list[Projection]
        List of projections to apply to the image.

    """
    # Create save directory
    if not os.path.exists(save_dir):
        os.makedirs(save_dir)

    # Select tif or tiff files in folder
    files = glob.glob(os.path.join(data_dir, "*.tif*"))
    # Iterate over files
    for file_index, file in enumerate(files):
        # Read image
        file_path = os.path.join(data_dir, file)
        image = TiffReader(
            file_path, project=projection, respect_initial_type=True
        ).get_processed_image()
        # Select random time steps to extract
        time_points_to_extract = np.random.choice(
            image.shape[time_dimension] - 1, frames_per_file, replace=False
        )
        time_points_to_extract = np.unique(time_points_to_extract).astype(int)
        # Iterate over time stamps
        selected_images = image.take(
            time_points_to_extract, axis=time_dimension
        )
        # Save frames
        for frame, selected_image in zip(
            time_points_to_extract, selected_images
        ):
            padded_frame = str(frame).zfill(
                3
            )  # so that cellpose take them in order
            selected_image = np.expand_dims(selected_image, axis=0)
            save_tiff(
                selected_image,
                f"{save_dir}/{padded_frame}_{Path(file).stem}.tif",
                original_order="TZCYX",
            )

        display_progress(
            "Generation in progress",
            file_index + 1,
            len(files),
        )


if __name__ == "__main__":
    DATA_DIR = ""
    SAVE_DIR = ""

    FRAMES_PER_FILE = 10  # how many frames to extract from each video
    TIME_DIMENSION = 0  # time dimension in the tiff file

    PROJECTION = [
        Projection(method=ProjectMethods.Channel, channels=[2], axis=1)
    ]  # channel 2 on dimension 1

    extract_frames(
        SAVE_DIR, DATA_DIR, TIME_DIMENSION, FRAMES_PER_FILE, PROJECTION
    )



import os
from typing import Optional

import numpy as np

from ..utils.cell_track import CellTrack
from ..factories.results_saving_factory import ResultsSavingFactory
from ..utils.mitosis_track import MitosisTrack


def perform_results_saving(
    exported_mitoses_dir: str,
    show: bool = False,
    save_dir: Optional[str] = None,
    verbose: bool = False,
    video: Optional[np.ndarray] = None,
    viewer: Optional["napari.Viewer"] = None,
    segmentation_results: Optional[np.ndarray] = None,
    cell_tracks: Optional[list[CellTrack]] = None,
) -> None:
    """Perform a series of tests, prints and plots following process.

    Parameters
    ----------
    video : np.ndarray
        Video. TYXC.
    exported_mitoses_dir : str
        Directory where mitosis tracks are saved.
    show : bool, optional
        Show plots, by default False.
    save_dir : Optional[str], optional
        Directory where to save results, by default None.
    verbose : bool, optional
        Verbose, by default False.
    video : Optional[np.ndarray], optional
        Video, by default None. Any dimension order.
    viewer : Optional["napari.Viewer"], optional
        Viewer, by default None.
    segmentation_results : Optional[np.ndarray], optional
        Segmentation results, by default None. TYX.
    """
    print("\n### RESULTS SAVING ###")

    # Create save_dir if specified and it does not exist
    if save_dir is not None and not os.path.exists(save_dir):
        os.makedirs(save_dir)

    mitosis_tracks: list[MitosisTrack] = []
    mitosis_video_names: list[str] = []
    # Iterate over "bin" files in exported_mitoses_dir
    for state_path in os.listdir(exported_mitoses_dir):
        with open(os.path.join(exported_mitoses_dir, state_path), "rb") as f:
            mitosis_track = MitosisTrack.load(f)
            mitosis_tracks.append(mitosis_track)
            mitosis_video_names.append(state_path.split("_mitosis_")[0])

    # Define lists and dictionaries to store results
    results_saving_factory = ResultsSavingFactory()
    results_saving_factory.update_cut_times(mitosis_tracks, verbose)

    # Protect against no detection
    if len(results_saving_factory.first_cut_times) == 0:
        print("No mitosis detected.")
        return

    # Perform a series of tests, prints and plots
    results_saving_factory.perform_t_test()
    results_saving_factory.print_analysis_summary(mitosis_tracks)
    results_saving_factory.save_csv_results(
        mitosis_tracks, mitosis_video_names, save_dir
    )
    results_saving_factory.box_plot_cut_differences(show, save_dir)
    results_saving_factory.plot_cut_distributions(show, save_dir)

    # Display in napari
    if video is not None:
        print("\nDisplaying results in Napari.")
        results_saving_factory.generate_napari_tracking_mask(
            mitosis_tracks,
            video,
            viewer,
            segmentation_results=segmentation_results,
            cell_tracks=cell_tracks,
        )



import torch

from cnn_framework.utils.data_loader_generators.classifier_data_loader_generator import (
    ClassifierDataLoaderGenerator,
)
from cnn_framework.utils.metrics.classification_accuracy import (
    ClassificationAccuracy,
)
from cnn_framework.utils.models.resnet_classifier import ResnetClassifier

from cut_detector.utils.mt_cut_detection.bridges_mt_model_manager import (
    BridgesMtModelManager,
)

from developers.binary_bridges_cnn.data_set_inference import BridgesCnnDataSet
from developers.binary_bridges_cnn.model_params import BinaryBridgesModelParams
from developers.binary_bridges_cnn.bridges_parser import BridgesParser
from developers.binary_bridges_cnn.tools import evaluate_frame_error


def main(params):
    loader_generator = ClassifierDataLoaderGenerator(params, BridgesCnnDataSet)
    _, _, test_dl = loader_generator.generate_data_loader()

    # Model definition
    model = ResnetClassifier(
        nb_classes=2,  # hard-coded from binary training
        nb_input_channels=len(params.c_indexes) * len(params.z_indexes),
        encoder_name=params.encoder_name,
    )
    model.load_state_dict(torch.load(params.model_load_path))

    manager = BridgesMtModelManager(model, params, ClassificationAccuracy)
    predictions = manager.predict(test_dl, nb_images_to_save=-1)
    manager.write_useful_information()

    evaluate_frame_error(
        predictions,
        test_dl.dataset.names,
        params.hmm_bridges_parameters_file,
    )


if __name__ == "__main__":
    parser = BridgesParser()
    args = parser.arguments_parser.parse_args()

    parameters = BinaryBridgesModelParams(inference_mode=True)
    parameters.update(args)

    main(parameters)



"""Playground to run micro-tubules cut detection on a single mitosis.
Predicted classes and bridge crops are displayed."""

import os
from typing import Optional
from matplotlib import pyplot as plt

from cnn_framework.utils.readers.tiff_reader import TiffReader

from cut_detector.data.tools import get_data_path
from cut_detector.models.tools import get_model_path
from cut_detector.factories.mt_cut_detection_factory import (
    MtCutDetectionFactory,
)
from cut_detector.utils.mitosis_track import MitosisTrack
from cut_detector.utils.tools import re_organize_channels


def main(
    image_path: Optional[str] = get_data_path("videos"),
    mitosis_path: Optional[str] = get_data_path("mitoses"),
    hmm_bridges_parameters_file: Optional[str] = os.path.join(
        get_model_path("hmm"), "hmm_bridges_parameters.npz"
    ),
    bridges_mt_cnn_model_path: Optional[str] = get_model_path(
        "bridges_mt_cnn"
    ),
    display_predictions_analysis=True,
    display_crops=True,
):
    """
    Parameters
    ----------
    image_path : str
        Path to the image to process.
    mitosis_path : str
        Path to the mitoses data.
    hmm_bridges_parameters_file : str
        Path to the HMM parameters file.
    bridges_mt_cnn_model_path : str
        Path to the bridges MT CNN model.
    display_predictions_analysis : bool
        If True, display the predicted classes before and after the HMM.
    display_crops : bool
        If True, display the crops of the bridges.
    """
    # If paths are directories, take their first file
    if os.path.isdir(image_path):
        image_path = os.path.join(image_path, os.listdir(image_path)[0])
    if os.path.isdir(mitosis_path):
        mitosis_path = os.path.join(mitosis_path, os.listdir(mitosis_path)[0])

    # Read data: image, mitosis_track
    image = TiffReader(image_path, respect_initial_type=True).image
    with open(mitosis_path, "rb") as f:
        mitosis_track = MitosisTrack.load(f)

    image = re_organize_channels(image.squeeze())  # TYXC

    factory = MtCutDetectionFactory()

    results = factory.update_mt_cut_detection(
        [mitosis_track],
        image,
        hmm_bridges_parameters_file,
        bridges_mt_cnn_model_path,
        debug_mode=True,
    )

    if display_predictions_analysis:
        plt.plot(results["predictions"][mitosis_track.id])
        if (
            mitosis_track.id in results["predictions_after_hmm"]
        ):  # if classification possible
            plt.plot(results["predictions_after_hmm"][mitosis_track.id])
        plt.title("Class bridges")
        plt.show()

    if display_crops:
        # Display series of crops
        for crop in results["images"][mitosis_track.id]:
            plt.figure()
            plt.imshow(crop[0], cmap="gray")
            plt.show()


if __name__ == "__main__":
    main()



import os
import random
from typing import Optional
from scipy.stats import ttest_1samp
import numpy as np
import matplotlib.pyplot as plt
from napari import Viewer
from tqdm import tqdm

from ..utils.cell_track import CellTrack, generate_tracking_movie
from ..utils.tools import re_organize_channels
from ..constants.tracking import TIME_RESOLUTION
from ..utils.mitosis_track import MitosisTrack
from ..utils.mt_cut_detection.impossible_detection import (
    ImpossibleDetection,
)


def grayscale_to_rgb(grayscale_image, channel_axis, minimum_value=50):
    """Convert grayscale image to RGB image.

    Parameters
    ----------
    grayscale_image: np.ndarray
        Grayscale image. TYX.
    channel_axis: int
        Channel axis.

    Returns
    -------
    np.ndarray
        RGB image.
    """
    grayscale_image = grayscale_image.astype(np.float32)  # TYX
    grayscale_image = grayscale_image / grayscale_image.max() * 255
    grayscale_image = np.clip(grayscale_image, 0, 255)
    # Avoid values close to 0 as they are too dark
    grayscale_image[grayscale_image > 0] = (
        grayscale_image[grayscale_image > 0] / 255 * (255 - minimum_value)
        + minimum_value
    )
    grayscale_image = grayscale_image.astype(np.uint8)  # TYX

    indexes = np.unique(grayscale_image)
    fake_second_channel = np.copy(grayscale_image)
    fake_third_channel = np.copy(grayscale_image)

    random_colors = get_random_different_colors(len(indexes), nb_channels=2)
    for random_color, index in zip(random_colors, indexes):
        if index == 0:
            continue
        fake_second_channel[grayscale_image == index] = random_color[0]
        fake_third_channel[grayscale_image == index] = random_color[1]

    grayscale_image = np.stack(
        [
            grayscale_image,
            fake_second_channel,
            fake_third_channel,
        ],
        axis=-1,
    )  # TYXC

    # Match original image shape
    grayscale_image = np.moveaxis(grayscale_image, 3, channel_axis)  # TCYX
    return grayscale_image


def get_random_different_colors(
    nb_colors: int, minimum_value=50, nb_channels=3, interval=15
) -> np.ndarray:
    """Get random different colors.

    Parameters
    ----------
    nb_colors: int
        Number of colors.
    minimum_value: int
        Minimum value for colors to avoid dark colors. The default is 50.
    nb_channels: int
        Number of channels. The default is 3.
    interval: int
        Interval between colors. The default is 15.

    Returns
    -------
    np.ndarray
        Colors. Shape: (nb_colors, nb_channels).
    """
    colors = []
    for _ in range(nb_channels):
        colors_pick_list = np.array(range(minimum_value, 256))
        channel_colors = []
        for _ in range(nb_colors):
            if len(colors_pick_list) == 0:
                colors_pick_list = np.array(range(minimum_value, 256))
            random_color = np.random.choice(colors_pick_list)
            channel_colors.append(random_color)
            # Remove values close to random_color to avoid similar colors
            colors_pick_list = np.delete(
                colors_pick_list,
                np.where(np.abs(colors_pick_list - random_color) < interval),
            )
        colors.append(channel_colors)

    colors = np.array(colors).T.astype(np.uint8)
    return np.array(colors)


class ResultsSavingFactory:
    """Factory to save results.

    Parameters
    ----------
    time_resolution: int
        Time resolution in minutes. The default is TIME_RESOLUTION.
    max_frame: float
        Maximum frame to consider. The default is np.inf.
    """

    def __init__(
        self,
        time_resolution=TIME_RESOLUTION,
        max_frame=np.inf,
    ):
        self.time_resolution = time_resolution
        self.max_frame = max_frame  # for debug, 48*6 <-> 48h * 6 frames/hour

        self.first_cut_times: list[int] = []  # time in absolute time
        self.first_cut_times_gt: list[int] = []  # time in absolute time
        self.cut_differences: list[int] = []

        self.mitosis_results_summary: dict[int, int] = {}
        self.gt_mitosis_results_summary: dict[int, int] = {}

        self.p_value = None

    def _print_weird_mitoses(
        self, selected_tracks: list[MitosisTrack], min_acceptable_frame: int
    ) -> None:
        """Print useful information about weird mitoses.

        Parameters
        ----------
        selected_tracks: list[MitosisTrack]
            List of selected mitosis tracks.
        min_acceptable_frame: int
            Minimum acceptable frame for a cut to be considered as early.
        """
        ordered_tracks = [
            track
            for track in selected_tracks
            if track.key_events_frame["first_mt_cut"]
            - track.key_events_frame["cytokinesis"]
            <= min_acceptable_frame
        ]
        ordered_tracks.sort(
            key=lambda x: x.key_events_frame["first_mt_cut"]
            - x.key_events_frame["cytokinesis"]
        )

        print("Weird mitoses (early cut):")
        for mitosis_track in ordered_tracks:
            print("")
            print(
                f"Track: {mitosis_track.id}_{mitosis_track.mother_track_id}_to_{','.join(str(daughter) for daughter in mitosis_track.daughter_track_ids)}"
            )
            print(f"Key events frame: {mitosis_track.key_events_frame}")
            print(
                "video frame cut",
                mitosis_track.key_events_frame["first_mt_cut"]
                - mitosis_track.min_frame
                + 1,
            )
            if mitosis_track.gt_key_events_frame is not None:
                print(
                    f"GT key events frame: {mitosis_track.gt_key_events_frame}"
                )
                print(
                    "video frame cut",
                    mitosis_track.gt_key_events_frame["first_mt_cut"]
                    - mitosis_track.min_frame
                    + 1,
                )

    def update_cut_times(
        self,
        mitosis_tracks: list[MitosisTrack],
        verbose: bool,
        min_acceptable_frame=13,
    ) -> None:
        """Update cut times and summary.

        Parameters
        ----------
        mitosis_tracks: list[MitosisTrack]
            List of mitosis tracks.
        verbose: bool
            Whether to print details about weird mitoses.
        min_acceptable_frame: int
            Minimum acceptable frame for a cut to be considered as early.
        """

        selected_tracks: list[MitosisTrack] = []  # kept tracks

        for mitosis_track in mitosis_tracks:
            # Get first cut frame and start of cytokinesis frame
            cyto_frame = mitosis_track.key_events_frame["cytokinesis"]
            cut_frame = mitosis_track.key_events_frame["first_mt_cut"]

            if cut_frame < 0 or cut_frame > self.max_frame:
                # No cut detected, for some reason
                cut_time = None
            else:
                # At least one cut was actually detected
                assert cut_frame >= cyto_frame  # should not be possible
                # Add the difference time to the list
                cut_time = (cut_frame - cyto_frame) * self.time_resolution
                self.first_cut_times.append(cut_time)
                selected_tracks.append(mitosis_track)

            # Update summary
            cut_id = min(cut_frame, 0)  # 0 if normal, negative otherwise

            if cut_id not in self.mitosis_results_summary:
                self.mitosis_results_summary[cut_id] = 0
            self.mitosis_results_summary[cut_id] += 1

            # Ignore if information is not there
            if (
                mitosis_track.gt_key_events_frame is None
                or "first_mt_cut" not in mitosis_track.gt_key_events_frame
            ):
                continue

            if (
                mitosis_track.gt_key_events_frame["first_mt_cut"]
                > self.max_frame
            ):
                # Ignore mitoses that are too late
                continue

            cut_time_gt = (
                mitosis_track.gt_key_events_frame["first_mt_cut"]
                - mitosis_track.gt_key_events_frame["cytokinesis"]
            ) * self.time_resolution
            self.first_cut_times_gt.append(cut_time_gt)

            if cut_time is not None:
                self.cut_differences.append(cut_time - cut_time_gt)

            # Update ground truth summary
            if cut_id not in self.gt_mitosis_results_summary:
                self.gt_mitosis_results_summary[cut_id] = 0
            self.gt_mitosis_results_summary[cut_id] += 1

        # Display few details on mitoses with very early cut
        if verbose:
            self._print_weird_mitoses(selected_tracks, min_acceptable_frame)

    def perform_t_test(self, alpha=0.05) -> None:
        """Compute t-test of the differences, which is supposed to be 0.

        Parameters
        ----------
        alpha: float
            Significance level. The default is 0.05.
        """
        if len(self.cut_differences) == 0:
            return None

        t_result = ttest_1samp(self.cut_differences, 0)
        self.p_value = t_result.pvalue

        # Output the results
        print(f"P-value: {self.p_value}")

        # Check if the result is statistically significant (e.g., using a significance level of 0.05)
        if self.p_value < alpha:
            print(
                "Reject the null hypothesis: There is a statistically significant difference."
            )
        else:
            print(
                "Fail to reject the null hypothesis: There is no statistically significant difference."
            )

    def print_analysis_summary(self, mitosis_tracks) -> None:
        """Print analysis summary.

        Parameters
        ----------
        mitosis_tracks: list[MitosisTrack]
            List of mitosis tracks.
        """
        print(f"Among the {len(mitosis_tracks)} mitoses detected, there are:")
        for cut_id, count in self.mitosis_results_summary.items():
            print(
                f"    - {count} mitoses in category {ImpossibleDetection(cut_id).name}"
            )

        if len(self.first_cut_times_gt) == 0:
            return

        print(
            f"Among the {len(self.first_cut_times_gt)} mitoses annotated, there are:"
        )
        for cut_id, count in self.gt_mitosis_results_summary.items():
            print(
                f"    - {count} mitoses in category {ImpossibleDetection(cut_id).name}"
            )

    def box_plot_cut_differences(
        self, show: bool, save_dir: Optional[str]
    ) -> None:
        """Plot box plot of cut differences.

        Parameters
        ----------
        show: bool
            Whether to show the plot.
        save_dir: Optional[str]
            Directory to save the plot.
        """
        if len(self.cut_differences) == 0:
            return

        plt.figure()
        plt.boxplot(self.cut_differences)
        plt.title("Box plot of cut differences")
        plt.ylabel("Cut difference (min)")

        if show:
            plt.show()

        if save_dir is not None:
            plt.savefig(os.path.join(save_dir, "first_mt_cut_differences.png"))

    def plot_cut_distributions(
        self,
        show: bool,
        save_dir: Optional[str],
    ) -> None:
        """Plot distribution of cut differences.

        Parameters
        ----------
        show: bool
            Whether to show the plot.
        save_dir: Optional[str]
            Directory to save the plot.
        """

        plt.figure()

        for cut_times, name in zip(
            [self.first_cut_times, self.first_cut_times_gt],
            ["Predicted", "Ground truth"],
        ):
            if len(cut_times) == 0:  # ignore empty gt list
                continue
            curve_values = []
            nb_values = len(cut_times)
            number_of_cut = 0
            for i in range(max(cut_times) + 1):
                count = cut_times.count(i)
                number_of_cut += count
                curve_values.append(number_of_cut / nb_values)
            plt.plot(curve_values, label=name)

        # Quartiles and median
        predicted_median = np.median(self.first_cut_times)
        gt_median = (
            np.median(self.first_cut_times_gt)
            if len(self.first_cut_times_gt)
            else None
        )
        predicted_q1 = np.percentile(self.first_cut_times, 25)
        predicted_q3 = np.percentile(self.first_cut_times, 75)
        plt.axvline(
            x=predicted_q1,
            color="r",
            linestyle="--",
            linewidth=0.5,
            label=f"Detected Q1: {predicted_q1}",
        )
        plt.axvline(
            x=predicted_median,
            color="r",
            linestyle="-",
            linewidth=0.5,
            label=f"Detected median: {predicted_median}"
            + (f"vs {gt_median} Ground truth median" if gt_median else ""),
        )
        plt.axvline(
            x=predicted_q3,
            color="r",
            linestyle="--",
            linewidth=0.5,
            label=f"Detected Q3: {predicted_q3}",
        )

        # add t test result to legend
        if self.p_value is not None:
            plt.legend(
                loc="lower right", title=f"t test p-value: {self.p_value:.2f}"
            )
        else:
            plt.legend(loc="lower right")

        plt.title(
            f"Proportion of MT cuts over time - {len(self.first_cut_times)} predicted mitoses"
            + (
                f"- {len(self.first_cut_times_gt)} ground truth mitoses"
                if len(self.first_cut_times_gt)
                else ""
            )
        )
        plt.xlabel("Time (min)")
        plt.ylabel("Proportion of first cut")

        if show:
            plt.show()

        if save_dir is not None:
            plt.savefig(
                os.path.join(save_dir, "first_mt_cut_distributions.png")
            )

    def save_csv_results(
        self,
        mitosis_tracks: list[MitosisTrack],
        mitosis_video_names: list[str],
        save_dir: Optional[str] = None,
    ) -> None:
        """Save results in CSV file.

        Parameters
        ----------
        mitosis_tracks: list[MitosisTrack]
            List of mitosis tracks.
        mitosis_video_names: list[str]
            List of mitosis video names.
        save_dir: Optional[str]
            Directory to save the results.
        """
        if save_dir is None:
            return

        csv_path = os.path.join(save_dir, "results.csv")
        # Create CSV results file if it does not exist
        if not os.path.exists(csv_path):
            with open(csv_path, "w") as f:
                f.write(
                    "video;id;mother track;daughter track(s);metaphase frame;cytokinesis frame;first MT cut frame;second MT cut frame;first MT cut time;second MT cut time\n"
                )
            f.close()

        def handle_impossible_detection(cut_frame: int) -> str:
            return (
                ImpossibleDetection(cut_frame).name
                if cut_frame < 0
                else str(cut_frame)
            )

        # Store useful results in global results file
        with open(csv_path, "a") as f:
            for mitosis_track, mitosis_video_name in zip(mitosis_tracks, mitosis_video_names):
                f.write(f"{mitosis_video_name};")
                f.write(f"{mitosis_track.id};")
                f.write(f"{mitosis_track.mother_track_id};")
                daughter_ids = ",".join(
                    [str(d) for d in mitosis_track.daughter_track_ids]
                )
                f.write(f"{daughter_ids};")
                f.write(f"{mitosis_track.key_events_frame['metaphase']};")
                cytokinesis_frame = mitosis_track.key_events_frame[
                    "cytokinesis"
                ]
                f.write(f"{cytokinesis_frame};")
                first_cut_frame = mitosis_track.key_events_frame[
                    "first_mt_cut"
                ]
                f.write(f"{handle_impossible_detection(first_cut_frame)};")
                second_cut_frame = mitosis_track.key_events_frame[
                    "second_mt_cut"
                ]
                f.write(f"{handle_impossible_detection(second_cut_frame)};")
                first_cut_time = (
                    (first_cut_frame - cytokinesis_frame)
                    * self.time_resolution
                    if first_cut_frame >= 0
                    else ""
                )
                f.write(f"{first_cut_time};")
                second_cut_time = (
                    (second_cut_frame - cytokinesis_frame)
                    * self.time_resolution
                    if second_cut_frame >= 0
                    else ""
                )
                f.write(f"{second_cut_time};\n")
        f.close()

    def generate_napari_tracking_mask(
        self,
        mitosis_tracks: list[MitosisTrack],
        video: np.ndarray,
        viewer: Optional[Viewer] = None,
        segmentation_results: Optional[np.ndarray] = None,
        cell_tracks: Optional[list[CellTrack]] = None,
    ) -> None:
        """Generate napari tracking mask.

        Parameters
        ----------
        mitosis_tracks: list[MitosisTrack]
            List of mitosis tracks.
        video: np.ndarray
            Video to process. Any dimension order.
        viewer: napari.Viewer
            Napari viewer.
        segmentation_results: np.ndarray
            Cellpose results. TYX.
        """

        channel_axis = np.argmin(video.shape)
        # By default, Napari assumes that images with 3 dimensions in last
        # axis are RGB
        rgb = channel_axis == 3

        # Re-organize channels
        video_to_process = re_organize_channels(video)  # TYXC

        # Video parameters
        nb_frames, height, width, _ = video_to_process.shape

        # Colors list
        colors = get_random_different_colors(len(mitosis_tracks))

        # Iterate over mitosis_tracks
        mitoses_results = np.zeros(
            (nb_frames, height, width, 3), dtype=np.uint8
        )  # TYXC
        for idx, mitosis_track in enumerate(tqdm(mitosis_tracks)):
            if not mitosis_track.display():
                continue
            _, mask_movie = mitosis_track.generate_video_movie(
                video_to_process
            )
            cell_indexes = np.where(mask_movie == 1)
            mask_movie = np.stack(
                [mask_movie, mask_movie, mask_movie], axis=-1
            )
            mask_movie[cell_indexes] = colors[idx]
            initial_mask = mitoses_results[
                mitosis_track.min_frame : mitosis_track.max_frame + 1,
                mitosis_track.position.min_y : mitosis_track.position.max_y,
                mitosis_track.position.min_x : mitosis_track.position.max_x,
                :,
            ]
            # Avoid colors overlap
            mitoses_results[
                mitosis_track.min_frame : mitosis_track.max_frame + 1,
                mitosis_track.position.min_y : mitosis_track.position.max_y,
                mitosis_track.position.min_x : mitosis_track.position.max_x,
                :,
            ] = np.maximum(mask_movie, initial_mask)
        # Match original image shape
        mitoses_results = np.moveaxis(mitoses_results, 3, channel_axis)  # TCYX
        assert mitoses_results.shape == video.shape

        # Use point + text instead of red point for mid_body
        points = []
        features = {"category": []}
        text = {
            "string": "{category}",
            "size": 10,
            "color": "white",
            "translation": np.array([-30, 0]),
        }
        for mitosis_track in tqdm(mitosis_tracks):
            if not mitosis_track.display():
                continue
            mid_body_legend = mitosis_track.get_mid_body_legend()
            for frame, frame_dict in mid_body_legend.items():
                single_layer_points = np.array(
                    [frame, frame_dict["y"], frame_dict["x"]]
                )
                if rgb:
                    points += [single_layer_points]
                    features["category"] += [frame_dict["category"]]
                else:
                    for idx in range(3):  # 3 channels
                        layer_points = np.insert(
                            single_layer_points, channel_axis, idx
                        )
                        points += [layer_points]
                        features["category"] += [frame_dict["category"]]
        features["category"] = np.array(features["category"])

        if viewer is None:
            return

        viewer.add_image(
            mitoses_results,
            name="Cell divisions",
            opacity=0.4,
            rgb=rgb,
            colormap=None if rgb else "inferno",
        )
        viewer.add_points(
            points,
            features=features,
            text=text,
            name="Mid-bodies",
        )

        if segmentation_results is not None:
            segmentation_results = grayscale_to_rgb(
                segmentation_results, channel_axis
            )
            assert segmentation_results.shape == video.shape

            viewer.add_image(
                segmentation_results,
                name="Segmentation",
                opacity=0.4,
                rgb=rgb,
                colormap=None if rgb else "inferno",
                visible=False,
            )

        if cell_tracks is not None:
            mitoses_results = generate_tracking_movie(
                cell_tracks, video_to_process
            )  # TYX

            mitoses_results = grayscale_to_rgb(mitoses_results, channel_axis)
            assert mitoses_results.shape == video.shape

            viewer.add_image(
                mitoses_results,
                name="Tracking",
                opacity=0.4,
                rgb=rgb,
                colormap=None if rgb else "inferno",
                visible=False,
            )



import os
import pickle
import ast
import csv
from typing import Optional
import torch
from torch.utils.data import DataLoader
import numpy as np

from cnn_framework.utils.tools import extract_patterns
from cnn_framework.utils.display_tools import display_progress
from cnn_framework.utils.model_managers.cnn_model_manager import (
    CnnModelManager,
)
from cnn_framework.utils.data_managers.default_data_manager import (
    DefaultDataManager,
)
from cnn_framework.utils.metrics.classification_accuracy import (
    ClassificationAccuracy,
)
from cnn_framework.utils.data_loader_generators.data_loader_generator import (
    collate_dataset_output,
)
from cnn_framework.utils.enum import PredictMode
from cnn_framework.utils.models.resnet_classifier import ResnetClassifier
from cnn_framework.utils.model_params.base_model_params import BaseModelParams

from .cnn_data_set import CnnDataSet
from .hidden_markov_models import HiddenMarkovModel


def re_organize_channels(image: np.ndarray) -> np.ndarray:
    """Expect a 4 dimensions image.
    Re-organize channels to get time as first dimension and channel at last.

    Parameters
    ----------
    image : np.ndarray
        Image to re-organize.

    Returns
    -------
    np.ndarray
        Re-organized image.
    """
    if image.ndim != 4:
        raise ValueError("Expect a 4 dimensions image.")

    # Get dimension index of smallest dimension, i.e. channels
    channels_dimension_index = np.argmin(image.shape)
    channels_dimension = image.shape[channels_dimension_index]
    if channels_dimension != 3:
        raise ValueError(
            "Expect 3 channels: SiR-tubulin/MKLP1/Phase contrast, in that order."
        )
    # Put channels at the back
    image = np.moveaxis(image, channels_dimension_index, 3)

    # Get dimension index of second smallest dimension, i.e. time
    second_dimension_index = np.argsort(image.shape)[1]
    # Put time at the front
    image = np.moveaxis(image, second_dimension_index, 0)

    return image


def get_annotation_file(
    video_path: str, mitosis_track, annotations_files: list[str]
) -> Optional[str]:
    """Get annotation file corresponding to a mitosis track. Return None if not found.

    Parameters
    ----------
    video_path : str
        Path to video file.
    mitosis_track : MitosisTrack
        Mitosis track to find annotation file for.
    annotations_files : list[str]
        List of annotation files.

    Returns
    -------
    Optional[str]
        Path to annotation file.
    """
    video_file = os.path.basename(video_path).split(".")[0]
    mitosis_id = f"mitosis_{mitosis_track.id}_"
    # Assume that mitosis id is consistent across all segmentation/tracking runs
    # (i.e. possible to use former annotations with new mitoses)
    annotation_path_candidates = extract_patterns(
        annotations_files, ["*" + video_file + "*" + mitosis_id + "*"]
    )
    # Check if video exists
    if len(annotation_path_candidates) == 0:
        return None

    return annotation_path_candidates[0]


def upload_annotations(
    annotations_folder: str,
    video_path: str,
    mitoses_folder: str,
    save: bool,
) -> tuple[int]:
    """Function used to upload annotations and perform mid-body detection evaluation.

    Parameters
    ----------
    annotations_folder : str
        Folder containing annotations.
    video_path : str
        Path to video file.
    mitoses_folder : str
        Folder containing mitosis tracks.
    save : bool
        Whether to update mitoses with annotations.

    Returns
    -------
    tuple[int]
        Number of mid-bodies detected and not detected.
    """

    # Read video
    video_name = os.path.basename(video_path).split(".")[0]

    mitosis_tracks = []
    # Iterate over "bin" files in exported_mitoses_dir
    for state_path in os.listdir(mitoses_folder):
        # Ignore if not for current video
        if video_name not in state_path:
            continue
        # Load mitosis track
        with open(os.path.join(mitoses_folder, state_path), "rb") as f:
            mitosis_track = pickle.load(f)

        # Add mitosis track to list
        mitosis_tracks.append(mitosis_track)

    # Get all files in annotations_folder
    annotations_files = []
    for root, _, files in os.walk(annotations_folder):
        for file in files:
            if ".xml" in file:
                annotations_files.append(os.path.join(root, file))

    # Used to evaluate mid-body detection
    mb_detected, mb_not_detected = 0, 0

    # Generate movie for each mitosis and save
    wrong_detections = []
    for i, mitosis_track in enumerate(mitosis_tracks):
        # If annotation is available, load it and update mitosis track
        annotation_file = get_annotation_file(
            video_path, mitosis_track, annotations_files
        )
        if annotation_file is not None:
            mitosis_track.update_mid_body_ground_truth(
                annotation_file, nb_channels=4
            )

            # Save updated mitosis track
            if save:
                daughter_track_ids = ",".join(
                    [str(d) for d in mitosis_track.daughter_track_ids]
                )
                state_path = f"{video_name}_mitosis_{mitosis_track.id}_{mitosis_track.mother_track_id}_to_{daughter_track_ids}.bin"
                save_path = os.path.join(
                    mitoses_folder,
                    state_path,
                )
                with open(save_path, "wb") as f:
                    pickle.dump(mitosis_track, f)

            # Evaluate mid-body detection (ignore triple divisions)
            if len(mitosis_track.daughter_track_ids) == 1:
                (
                    is_correctly_detected,
                    percent_detected,
                    average_position_difference,
                ) = mitosis_track.evaluate_mid_body_detection()
                if is_correctly_detected:
                    mb_detected += 1
                else:
                    mb_not_detected += 1
                    wrong_detections.append(
                        {
                            "path": state_path,
                            "percent_detected": percent_detected,
                            "average_position_difference": average_position_difference,
                        }
                    )
                    mitosis_track.evaluate_mid_body_detection()

        display_progress(
            "Update with annotations and evaluate",
            i + 1,
            len(mitosis_tracks),
            additional_message=f"Image {i+1}/{len(mitosis_tracks)}",
        )

    # Print wrong detections
    if len(wrong_detections) > 0:
        print("\nWrong detections:")
        for wrong_detection in wrong_detections:
            print(
                f"{wrong_detection['path']}: detected {wrong_detection['percent_detected']}% with avg distance {wrong_detection['average_position_difference']}"
            )

    if (mb_detected + mb_not_detected) == 0:
        print("No mid-body detection evaluation possible.")
        return 0, 0

    print(
        f"\nMid-body detection evaluation: {mb_detected / (mb_detected + mb_not_detected) * 100:.2f}%"
    )

    return mb_detected, mb_not_detected


def upload_annotations_folder(
    annotations_folder: str,
    video_folder: str,
    mitoses_folder: str,
    save: bool,
) -> tuple[int]:
    """Main annotations upload function.

    Parameters
    ----------
    annotations_folder : str
        Folder containing annotations.
    video_path : str
        Path to video folder.
    mitoses_folder : str
        Folder containing mitosis tracks.
    save : Optional[bool], optional
        Whether to update mitoses with annotations, by default True.

    Returns
    -------
    tuple[int]
        Number of mid-bodies detected and not detected.
    """
    all_detected, all_not_detected = 0, 0

    video_paths = os.listdir(video_folder)
    for video_idx, video in enumerate(video_paths):
        video_progress = f"{video_idx + 1}/{len(video_paths)}"
        print(f"\nVideo {video_progress}...")
        # Perform mid-body detection evaluation
        detected, not_detected = upload_annotations(
            annotations_folder,
            os.path.join(video_folder, video),
            mitoses_folder,
            save=save,
        )
        all_detected += detected
        all_not_detected += not_detected

    return all_detected, all_not_detected


def csv_parameters_to_dict(parameters_path: str) -> dict:
    """Read parameters from a csv file and return them as a dictionary.
    Assume to skip some lines if they do not have the right format.

    Parameters
    ----------
    parameters_path : str
        Path to csv file.

    Returns
    -------
    dict
        Parameters as dictionary.
    """
    result_dict = {}
    with open(parameters_path, "r") as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=",")
        for row in csv_reader:
            splitted_row = row[0].split(";", 1)
            if len(splitted_row) == 1:  # if there is no data in second column
                continue
            if len(row) > 1:  # if there is a comma in second column
                splitted_row[1] += "," + ",".join(row[1:])
            key, value = splitted_row[0], splitted_row[1]
            if (
                len(value) and value[0] == "[" and value[-1] == "]"
            ):  # transform string to list
                value = ast.literal_eval(value)
            result_dict[key] = value
    return result_dict


def perform_cnn_inference(
    model_path: str,
    images: list[np.array],
    cnn_model_params: BaseModelParams,
    cnn_data_set=CnnDataSet,
    cnn_classifier=ResnetClassifier,
    model_manager=CnnModelManager,
) -> list[int]:
    """Perform CNN inference on a list of images.

    Parameters
    ----------
    model_path : str
        CNN model path.
    images :  list[np.array]
        list[CYX]
    cnn_model_params : BaseModelParams
        Model parameters.
    cnn_data_set : AbstractDataSet
        Dataset to read data.
    cnn_classifier : ResnetClassifier
        CNN classifier.
    model_manager : CnnModelManager
        Model manager.

    Returns
    -------
    predictions : list[int]
        Predicted classes.

    """
    # Metaphase model parameters
    model_parameters = cnn_model_params()
    # Modify parameters for inference
    model_parameters.train_ratio = 0
    model_parameters.val_ratio = 0
    model_parameters.test_ratio = 1
    model_parameters.models_folder = model_path

    # Read csv parameters file
    parameters_file = os.path.join(model_path, "parameters.csv")
    training_parameters = csv_parameters_to_dict(parameters_file)

    # Load pretrained model
    model = cnn_classifier(
        nb_classes=int(training_parameters["nb_classes"]),
        nb_input_channels=len(training_parameters["c_indexes"])
        * len(training_parameters["z_indexes"]),
        encoder_name=training_parameters["encoder_name"],
    )

    map_location = None
    if not torch.cuda.is_available():
        map_location = torch.device("cpu")
        print("No GPU found, using CPU.")
    model.load_state_dict(
        torch.load(
            os.path.join(model_path, f"{model_parameters.name}.pt"),
            map_location=map_location,
        )
    )

    # Test (no sampler to keep order)
    dataset_test = cnn_data_set(
        images,
        is_train=False,
        names=[f"{idx}.ext" for idx in range(len(images))],
        data_manager=DefaultDataManager(),
        params=model_parameters,
    )
    test_dl = DataLoader(
        dataset_test,
        batch_size=model_parameters.batch_size,
        collate_fn=collate_dataset_output,
    )

    manager = model_manager(model, model_parameters, ClassificationAccuracy)

    predictions = manager.predict(
        test_dl,
        predict_mode=PredictMode.GetPrediction,
        nb_images_to_save=0,
    )  # careful, this is scores and not probabilities
    predictions = [int(np.argmax(p)) for p in predictions]

    return predictions


def apply_hmm(hmm_parameters, sequence) -> list[int]:
    """Correct the sequence of classes using HMM.

    Parameters
    ----------
    hmm_parameters : dict
        HMM parameters.
    sequence : list[int]
        Sequence of classes.

    Returns
    -------
    list[int]
        Corrected sequence of classes.
    """
    # Define observation sequence
    obs_seq = np.asarray(sequence, dtype=np.int32)

    # Define HMM model & run inference
    model = HiddenMarkovModel(
        hmm_parameters["A"], hmm_parameters["B"], hmm_parameters["pi"]
    )
    states_seq, _ = model.viterbi_inference(obs_seq)

    return states_seq


def get_video_path(video_name: str, mitoses_path: str) -> str:
    """Return the path of the video corresponding to the given name.
    NB: used for developers only.

    Parameters
    ----------
    video_name : str
        Name of the video.
    mitoses_path : str
        Path to the mitoses folder.

    Returns
    -------
    str
        Path of the video.
    """
    # Video name start (acquisition name)
    video_name_start = "*" + video_name.split("mitosis")[0].split(" ")[-1]
    # Video name end (x_to_y.tiff)
    video_name_end = "*" + "_".join(video_name.split("_")[-3:])
    video_name_candidates = extract_patterns(
        os.listdir(mitoses_path), [video_name_start + video_name_end]
    )
    # Check if video exists
    if len(video_name_candidates) == 0:
        return None

    # Construct video path
    video_path = os.path.join(mitoses_path, video_name_candidates[0])

    return video_path



"""Image tools."""

import operator
from typing import Optional
import numpy as np


def crop_center(img: np.ndarray, bounding: tuple[int]) -> np.ndarray:
    """Crop image to bounding box.

    Parameters
    ----------
    img : np.ndarray
        Image to crop.
    bounding : tuple
        Bounding box to crop to.

    Returns
    -------
    np.ndarray
        Cropped image.
    """
    start = tuple(
        map(lambda a, da: (a - da + 1) // 2, img.shape, bounding)
    )  # 1 is used to match torch method
    end = tuple(map(operator.add, start, bounding))
    slices = tuple(map(slice, start, end))
    return img[slices]


def get_padding(
    input_shape: tuple[int], output_shape: tuple[int]
) -> tuple[int]:
    """Get padding to apply to input_shape to get output_shape.

    Parameters
    ----------
    input_shape : tuple
        Shape of the input image.
    output_shape : tuple
        Shape of the output image.

    Returns
    -------
    tuple[int]
        Padding to apply to input_shape to get output_shape.
    """
    (_, height, width) = input_shape
    (_, desired_height, desired_width) = output_shape

    # Define margins for both sides
    pad_margin_w_left = abs(desired_width - width) // 2
    pad_margin_w_right = abs(desired_width - width) - pad_margin_w_left

    pad_margin_h_top = abs(desired_height - height) // 2
    pad_margin_h_bottom = abs(desired_height - height) - pad_margin_h_top

    return (
        pad_margin_w_left,
        pad_margin_h_top,
        pad_margin_w_right,
        pad_margin_h_bottom,
    )


def resize_padding(
    image: np.ndarray,
    output_shape: tuple[int],
    mode: str,
    pad_margin_w: Optional[int],
    pad_margin_h: Optional[int],
) -> np.ndarray:
    """Resize image to output_shape with padding.

    Parameters
    ----------
    image : np.ndarray
        Image to resize.
    output_shape : tuple
        Shape of the output image.
    mode : str
        Mode to use for padding. "zero" or "min".
    pad_margin_w : Optional[int]
        Padding margin for width.
    pad_margin_h : Optional[int]
        Padding margin for height.

    Returns
    -------
    np.ndarray
        Resized image with padding.
    """
    channels_to_stack = []
    (depth, height, width) = image.shape
    (_, desired_height, desired_width) = output_shape

    # Crop image if it is too big
    if height > desired_height:
        image = crop_center(image, (depth, desired_height, width))
        height = desired_height
    if width > desired_width:
        image = crop_center(image, (depth, height, desired_width))
        width = desired_width

    raw_padding = get_padding(
        image.shape, output_shape
    )  # left, top, right and bottom

    # Define margins for both sides
    if pad_margin_w is None:
        pad_margin_w_left = raw_padding[0]
        pad_margin_w_right = raw_padding[2]
    else:
        pad_margin_w_left, pad_margin_w_right = pad_margin_w

    if pad_margin_h is None:
        pad_margin_h_top = raw_padding[1]
        pad_margin_h_bottom = raw_padding[3]
    else:
        pad_margin_h_top, pad_margin_h_bottom = pad_margin_h

    # Define padding
    padding = (
        (pad_margin_h_top, pad_margin_h_bottom),
        (pad_margin_w_left, pad_margin_w_right),
    )

    # Pad channels with minimal value of current channel
    for channel_image in range(depth):
        if mode == "min":
            constant_values = np.min(image[channel_image, ...])
        elif mode == "zero":
            constant_values = 0
        else:
            raise ValueError("Unknown mode")
        channel_image = np.pad(
            image[channel_image, ...],
            padding,
            mode="constant",
            constant_values=constant_values,
        )
        channels_to_stack.append(channel_image)

    # Stack padded channels and return
    return np.stack(channels_to_stack)


def resize_image(
    image,
    output_shape: Optional[tuple[int]] = None,
    method="zero",
    pad_margin_w: Optional[int] = None,
    pad_margin_h: Optional[int] = None,
):
    """Resize image to output_shape with padding. Expect image to be a 3D numpy array CYX.
    Output_shape is a tuple CYX.

    Parameters
    ----------
    image : np.ndarray
        Image to resize.
    output_shape : Optional[tuple]
        Shape of the output image.
    method : str
        Method to use for padding. "zero" or "min".
    pad_margin_w : Optional[int]
        Padding margin for width.
    pad_margin_h : Optional[int]
        Padding margin for height.

    Returns
    -------
    np.ndarray
        Resized image with padding.
    """

    if method in ["min", "zero"]:
        # Compute output_shape
        if output_shape is None:
            if pad_margin_h is None or pad_margin_w is None:
                raise ValueError(
                    "If output_shape is None, pad_margin_h and pad_margin_w must be specified"
                )
            output_shape = (
                None,
                sum(pad_margin_h) + image.shape[1],
                sum(pad_margin_w) + image.shape[2],
            )
        # Protect against output_shape and margins inconsistency
        if output_shape is not None and pad_margin_h is not None:
            assert sum(pad_margin_h) == output_shape[1] - image.shape[1]
        if output_shape is not None and pad_margin_w is not None:
            assert sum(pad_margin_w) == output_shape[2] - image.shape[2]
        # Pad to required_dimensions
        return resize_padding(
            image, output_shape, method, pad_margin_w, pad_margin_h
        )

    raise ValueError(f"Unknown resize method: {method}")


def smart_cropping(
    image: np.ndarray,
    margin: int,
    min_x: int,
    min_y: int,
    max_x: Optional[int] = None,
    max_y: Optional[int] = None,
    fade_margin=False,
    pad=False,
) -> np.ndarray:
    """
    Crop microscopy image with smart margin.

    Parameters
    ----------
    image : np.ndarray
        Image to crop. Shape must be (..., H, W).
    margin : int
        Margin to apply.
    min_x : int
        Minimum x coordinate.
    min_y : int
        Minimum y coordinate.
    max_x : Optional[int]
        Maximum x coordinate.
    max_y : Optional[int]
        Maximum y coordinate.
    fade_margin : bool
        If True, fade margin.
    pad : bool
        If True, pad image.

    Returns
    -------
    np.ndarray
        Cropped image.
    """

    # Get image shape
    height, width = image.shape[-2:]

    # Define empty rectangle if single point is given
    if max_x is None:
        max_x = min_x
    if max_y is None:
        max_y = min_y

    # Get smart margin
    clipped_min_y = max(min_y - margin, 0)
    clipped_max_y = min(max_y + margin, height)
    clipped_min_x = max(min_x - margin, 0)
    clipped_max_x = min(max_x + margin, width)

    # Crop image
    clipped_image = np.copy(
        image[..., clipped_min_y:clipped_max_y, clipped_min_x:clipped_max_x]
    )

    # Fade margin
    if fade_margin:
        margin_positions = (
            [
                [y, x]
                for x in range(min_x - clipped_min_x)
                for y in range(clipped_max_y - clipped_min_y)
            ]  # left margin
            + [
                [y, x + max_x - clipped_min_x]
                for x in range(clipped_max_x - max_x)
                for y in range(clipped_max_y - clipped_min_y)
            ]  # right margin
            + [
                [y, x]
                for y in range(min_y - clipped_min_y)
                for x in range(min_x - clipped_min_x, max_x - clipped_min_x)
            ]  # top margin
            + [
                [y + max_y - clipped_min_y, x]
                for y in range(clipped_max_y - max_y)
                for x in range(min_x - clipped_min_x, max_x - clipped_min_x)
            ]  # bottom margin
        )
        # Fade margin
        for margin_position in margin_positions:
            clipped_image[..., margin_position[0], margin_position[1]] = (
                clipped_image[..., margin_position[0], margin_position[1]] / 2
            )

    # Pad if necessary
    if pad and clipped_image.shape[-1] * clipped_image.shape[-2] != (
        max_x - min_x + 2 * margin
    ) * (max_y - min_y + 2 * margin):
        # Compute padding
        pad_margin_h = (
            clipped_min_y - (min_y - margin),
            max_y + margin - clipped_max_y,
        )
        pad_margin_w = (
            clipped_min_x - (min_x - margin),
            max_x + margin - clipped_max_x,
        )
        # Pad
        clipped_image = resize_image(
            clipped_image,
            pad_margin_h=pad_margin_h,
            pad_margin_w=pad_margin_w,
            method="zero",
        )

    return clipped_image


def cell_counter_frame_to_video_frame(
    cell_counter_frame: int, nb_channels=4
) -> int:
    """
    Cell counter index starts at 1, just like Fiji.

    To count frame, it just concatenates all channels.
    For example, with 4 channels, frames 1, 2, 3 and 4 will be frame 1,
    frames 5, 6, 7 and 8 will be frame 2, etc.
    """
    return (cell_counter_frame - 1) // nb_channels



import os
from matplotlib import pyplot as plt
from cnn_framework.utils.readers.tiff_reader import TiffReader


def annotate_cells(
    original_folder: str,
    class_folders: list[str],
    max_number_of_files: int,
    need_last_cells: bool,
) -> None:
    """Annotate cells in images.

    Parameters
    ----------
    original_folder : str
        Path to the folder containing the images to annotate.
    class_folders : list[str]
        List of paths to the folders where to save the annotated images.
        For each image, if you type "0" then the image will be stored in the first of these folders.
        Typically: 0 for interphase, 1 for metaphase and 2 for death.
    max_number_of_files : int
        Maximum number of files in class folder.
    need_last_cells : bool
        Whether to annotate the last cells first.
        Used to get more annotations of dead cells, which are less common than the others
        and likely to be at the end of the video.

    """
    # Create folders to save results if they do not exist yet
    for class_folder in class_folders:
        os.makedirs(class_folder, exist_ok=True)
    full_folders = [False for _ in range(len(class_folders))]

    files = os.listdir(original_folder)

    if need_last_cells:
        files.sort(key=lambda x: -int(x.split("_")[-1].split(".")[0]))

    for file_name in files:
        # If all folders have enough images, break
        if full_folders.count(True) == len(full_folders):
            break

        file_path = os.path.join(original_folder, file_name)
        image = TiffReader(file_path).get_processed_image()
        image = image.squeeze()

        # Function to handle key pressing and writing file in .txt accordingly
        def quit_figure(event):
            category = int(event.key)
            class_directory = class_folders[category]

            # Delete file if key is 9 or if class folder is full
            if category == 9 or full_folders[category] is True:
                os.remove(file_path)
            # If class folder is not full, move file to it
            else:
                os.rename(file_path, os.path.join(class_directory, file_name))
                if len(os.listdir(class_directory)) >= max_number_of_files:
                    full_folders[category] = True
            plt.close(event.canvas.figure)

        # Plot image and connect key press event
        nb_channels = image.shape[-1]

        for idx in range(nb_channels):
            plt.subplot(1, nb_channels, idx + 1)
            if idx == 1:
                plt.title(file_name)
            plt.imshow(image[..., idx], cmap="gray")

        plt.gcf().canvas.mpl_connect("key_press_event", quit_figure)

        # Display on whole screen
        figManager = plt.get_current_fig_manager()
        figManager.window.showMaximized()

        plt.show()


if __name__ == "__main__":
    # Parameters
    ORIGINAL_FOLDER = ""
    CLASS_FOLDERS = ["", "", ""]
    MAX_NUMBER_OF_FILES = 100  # Maximum number of files in class folder

    annotate_cells(
        ORIGINAL_FOLDER,
        CLASS_FOLDERS,
        MAX_NUMBER_OF_FILES,
        need_last_cells=False,
    )



from __future__ import annotations

from typing import Optional

from .spot import Spot


class MidBodySpot(Spot):
    """Mid-body spot.

    Parameters
    ----------
    frame : int
        Frame of the spot.
    x : int
        X coordinate of the spot.
    y : int
        Y coordinate of the spot.
    intensity : Optional[float], optional
        Intensity of the spot, by default None.
    sir_intensity : Optional[float], optional
        Sir intensity of the spot, by default None.
    area : Optional[float], optional
        Area of the spot, by default None.
    circularity : Optional[float], optional
        Circularity of the spot, by default None.
    """

    def __init__(
        self,
        frame: int,
        x: int,
        y: int,
        intensity: Optional[float] = None,
        sir_intensity: Optional[float] = None,
        area: Optional[float] = None,
        circularity: Optional[float] = None,
    ):
        super().__init__(frame, x, y)
        self.intensity = intensity
        self.sir_intensity = sir_intensity
        self.area = area
        self.circularity = circularity

        self.parent_spot: Optional[MidBodySpot] = None
        self.child_spot: Optional[MidBodySpot] = None

        self.track_id: Optional[int] = None

    @staticmethod
    def get_extra_features_name() -> list[str]:
        """Get the extra features name for tracking.

        Returns
        -------
        list[str]
            List of extra features name.
        """
        return ["mklp_intensity", "sir_intensity"]

    def get_extra_coordinates(self) -> list[float | None]:
        """Get the extra coordinates for tracking.

        Returns
        -------
        list[float | None]
            List of extra coordinates.
        """
        return [self.intensity, self.sir_intensity]



from __future__ import annotations
from ..constants.tracking import FRAMES_AROUND_METAPHASE


class MetaphaseSequence:
    """Class representing a sequence of metaphase frames.

    Parameters
    ----------
    sequence : list[int]
        Sequence of metaphase absolute frames.
    track_id : int
        Track id.
    """

    def __init__(self, sequence: list[int], track_id: int):
        self.first_frame = min(sequence)
        self.last_frame = max(sequence)

        self.track_id = track_id

    def is_mother_candidate(self, frame):
        """Check if the frame can be the first one of a daughter frame form this metaphase.
        Returns True if the frame is close to the last frame of the metaphase.

        Parameters
        ----------
        frame : int
            Frame number.

        Returns
        -------
        bool
            Whether the frame is a daughter candidate.
        """
        return (
            abs(self.last_frame - frame) < FRAMES_AROUND_METAPHASE
            and frame >= self.first_frame
        )

    def is_same(self, metaphase_sequence: MetaphaseSequence):
        """Check if the metaphase sequences are the same.

        Parameters
        ----------
        metaphase_sequence : MetaphaseSequence
            Metaphase sequence.

        Returns
        -------
        bool
            Whether the metaphase sequences are the same.
        """
        return (
            self.track_id == metaphase_sequence.track_id
            and self.last_frame == metaphase_sequence.last_frame
        )



from skimage.measure import find_contours
import numpy as np

from ..cell_spot import CellSpot


def get_spots_from_frame(
    frame: int, cellpose_result: np.ndarray
) -> tuple[int, list[CellSpot]]:
    """Extract spots from a single frame.

    Parameters
    ----------
    frame : int
        Frame number.
    cellpose_result : np.ndarray
        TYX

    Returns
    -------
    list[CellSpot]
        List of cell spots.
    """
    # Pad cellpose results to ensure that the contours are closed
    max_cellpose = np.max(cellpose_result)
    padded_cellpose_result = np.pad(
        cellpose_result, pad_width=1, mode="constant"
    )

    cell_spots = []
    for i in range(1, max_cellpose + 1):
        contours = find_contours(padded_cellpose_result == i)
        sorted_contours = sorted(
            contours,
            key=lambda contour: contour.shape[0],
            reverse=True,
        )
        # -1 to remove padding
        list_y = [int(l[0] - 1) for l in sorted_contours[0]]
        list_x = [int(l[1] - 1) for l in sorted_contours[0]]
        abs_min_x, abs_max_x, abs_min_y, abs_max_y = (
            np.abs(np.min(list_x)),
            np.abs(np.max(list_x)),
            np.abs(np.min(list_y)),
            np.abs(np.max(list_y)),
        )
        # Compute cell centroid
        cell_centroid = centroid(
            list_y,
            list_x,
        )  # (y, x)
        cell_spot = CellSpot(
            frame,
            cell_centroid[1],  # x
            cell_centroid[0],  # y
            -1,
            abs_min_x,
            abs_max_x,
            abs_min_y,
            abs_max_y,
            [[x, y] for x, y in zip(list_x, list_y)],
        )
        cell_spots.append(cell_spot)

    return frame, cell_spots


def signed_area(x: list[float], y: list[float]) -> float:
    """Compute the signed area of a polygon.

    Parameters
    ----------
    x : list[float]
        List of x-coordinates of the polygon.
    y : list[float]
        List of y-coordinates of the polygon.

    Returns
    -------
    float
        Signed area of the polygon.
    """
    n = len(x)
    a = 0.0
    for i in range(n - 1):
        a += x[i] * y[i + 1] - x[i + 1] * y[i]
    return (a + x[n - 1] * y[0] - x[0] * y[n - 1]) / 2.0


def centroid(x: list[float], y: list[float]) -> list[float]:
    """Calculate the centroid of a polygon.

    Parameters
    ----------
    x : list[float]
        List of x-coordinates of the polygon.
    y : list[float]
        List of y-coordinates of the polygon.

    Returns
    -------
    list[float]
        Coordinates of the centroid of the polygon.
    """
    area = signed_area(x, y)
    ax = 0.0
    ay = 0.0
    n = len(x)
    for i in range(n - 1):
        w = x[i] * y[i + 1] - x[i + 1] * y[i]
        ax += (x[i] + x[i + 1]) * w
        ay += (y[i] + y[i + 1]) * w

    w0 = x[n - 1] * y[0] - x[0] * y[n - 1]
    ax += (x[n - 1] + x[0]) * w0
    ay += (y[n - 1] + y[0]) * w0

    return [int(ax / 6.0 / area), int(ay / 6.0 / area)]



import os
import pickle
from typing import Optional
import numpy as np

from ..factories.mt_cut_detection_factory import MtCutDetectionFactory
from ..models.tools import get_model_path
from ..utils.mitosis_track import MitosisTrack


def perform_mt_cut_detection(
    raw_video: np.ndarray,
    video_name: str,
    exported_mitoses_dir: str,
    hmm_bridges_parameters_file: Optional[str] = os.path.join(
        get_model_path("hmm"), "hmm_bridges_parameters.npz"
    ),
    bridges_mt_cnn_model_path: Optional[str] = get_model_path(
        "bridges_mt_cnn"
    ),
    save: bool = True,
) -> list[MitosisTrack]:
    """Perform micro-tubules cut detection.

    Parameters
    ----------
    raw_video : np.ndarray
        Raw video. TYXC.
    video_name : str
        Video name.
    exported_mitoses_dir : str
        Directory where mitosis tracks are saved.
    hmm_bridges_parameters_file : Optional[str], optional
        HMM bridges parameters file.
    bridges_mt_cnn_model_path : Optional[str], optional
        Bridges micro-tubules CNN model path.
    save : bool, optional
        Save updated mitosis tracks, by default True.
    """
    print("\n### MICRO-TUBULES CUT DETECTION ###")

    mitosis_tracks: list[MitosisTrack] = []
    # Iterate over "bin" files in exported_mitoses_dir
    for state_path in os.listdir(exported_mitoses_dir):
        # Ignore if not for current video
        if video_name not in state_path:
            continue
        # Load mitosis track
        with open(os.path.join(exported_mitoses_dir, state_path), "rb") as f:
            mitosis_track = MitosisTrack.load(f)

        # Add mitosis track to list
        mitosis_tracks.append(mitosis_track)

    # Perform cut detection
    mt_cut_detector = MtCutDetectionFactory()
    mt_cut_detector.update_mt_cut_detection(
        mitosis_tracks,
        raw_video,
        hmm_bridges_parameters_file,
        bridges_mt_cnn_model_path,
    )

    # Save updated mitosis tracks
    if save:
        for mitosis_track in mitosis_tracks:
            daughter_track_ids = ",".join(
                [str(d) for d in mitosis_track.daughter_track_ids]
            )
            state_path = f"{video_name}_mitosis_{mitosis_track.id}_{mitosis_track.mother_track_id}_to_{daughter_track_ids}.bin"
            save_path = os.path.join(
                exported_mitoses_dir,
                state_path,
            )
            with open(save_path, "wb") as f:
                pickle.dump(mitosis_track, f)

    return mitosis_tracks



import os
from skimage import io

from cut_detector._widget import mid_body_detection
from cut_detector.data.tools import get_data_path
from cut_detector.widget_functions.mid_body_detection import (
    perform_mid_body_detection,
)


def test_open_mid_body_detection_widget():
    """Test opening the mid body detection widget."""
    mid_body_detection()


def test_mid_body_detection():
    """Test mid body detection."""
    video = io.imread(
        os.path.join(get_data_path("videos"), "example_video.tif")
    )  # TYXC

    # Run normal process
    mitosis_tracks = perform_mid_body_detection(
        video,
        "example_video",
        get_data_path("mitoses"),
        get_data_path("tracks"),
        save=False,
        parallel_detection=True,
    )

    is_correctly_detected, _, _ = mitosis_tracks[
        0
    ].evaluate_mid_body_detection()
    assert is_correctly_detected

    # Perform other strategies
    perform_mid_body_detection(
        video,
        "example_video",
        get_data_path("mitoses"),
        get_data_path("tracks"),
        save=False,
        parallel_detection=False,
    )
    perform_mid_body_detection(
        video,
        "example_video",
        get_data_path("mitoses"),
        get_data_path("tracks"),
        save=False,
        parallel_detection=False,
        detection_method="h_maxima",
    )



from __future__ import annotations

from abc import ABC, abstractmethod
import numpy as np


class Spot(ABC):
    """Class used for both cell and mid-body spots.

    Parameters
    ----------
    frame : int
        Frame of the spot.
    x : int
        X coordinate of the spot.
    y : int
        Y coordinate of the spot.
    """

    def __init__(self, frame: int, x: int, y: int):
        self.frame = frame
        self.x = x
        self.y = y

    def get_position(self) -> np.ndarray:
        """Return position as numpy array.

        Returns
        -------
        np.ndarray
            Position as numpy array.
        """
        return np.array([self.x, self.y])

    def distance_to(self, other_spot: Spot) -> float:
        """Compute distance between two spots.

        Parameters
        ----------
        other_spot : Spot
            Other spot to compute distance to.

        Returns
        -------
        float
            Distance between two spots.
        """
        return np.linalg.norm(
            np.array(self.get_position()) - np.array(other_spot.get_position())
        )

    @staticmethod
    @abstractmethod
    def get_extra_features_name() -> list[str]:
        """To have better tracking for classes that inherit from this one.
        These classes can declare here their extra coordinate fields.

        Example:
        If you have extra features, like 'color' and 'shape', implement this
        the following way:
        ```python
        def get_extra_features_name() -> list[str]:
            return ["color", "shape"]
        ```
        """
        return []

    @abstractmethod
    def get_extra_coordinates(self) -> list:
        """To have better tracking for classes that inherit from this one.
        These classes can return here their extra coordinate fields.

        Example:
        If you have extra features, like 'color' and 'shape', implement this
        the following way:
        ```python
        def get_extra_features_name(self) -> list[str]:
            return [self.color, self.shape]
        ```
        """
        return []

    def __str__(self) -> str:
        return f"x:{self.x} y:{self.y}"



from cut_detector.data.tools import get_data_path
from cut_detector.utils.tools import upload_annotations_folder


def test_annotations_upload():
    """Test manual CellCounter annotations upload."""

    annotations_folder = get_data_path("annotations")
    video_folder = get_data_path("videos")
    mitoses_folder = get_data_path("mitoses")

    _, not_detected = upload_annotations_folder(
        annotations_folder, video_folder, mitoses_folder, save=False
    )

    assert not_detected == 0



import numpy as np
from PIL import Image, ImageDraw

from .box_dimensions import BoxDimensions


class BoxDimensionsContour(BoxDimensions):
    """A class to store the dimensions of a box and the precise contour."""

    def __init__(self):
        super().__init__()
        self.list_points = []  # (x, y)

    def get_mask(self, local_shape: tuple[int]) -> np.ndarray:
        """Generate spot mask.

        Parameters
        ----------
        local_shape : tuple
            Shape of the mask.

        Returns
        -------
        np.ndarray
            Mask 2D.
        """
        mask = np.zeros(local_shape)
        for points in self.list_points:
            binary_image = Image.new("1", (local_shape[1], local_shape[0]), 0)
            draw = ImageDraw.Draw(binary_image)
            draw.polygon(
                [(point[1], point[0]) for point in points],
                outline=1,
                fill=1,
            )
            mask = np.logical_or(mask, np.array(binary_image))
        mask = mask.astype(np.uint8)
        return mask

    def update_list_points(self, relative: bool) -> None:
        """Update list of contour points.

        Parameters
        ----------
        relative : bool
            Relative or absolute coordinates.
        """
        # Switch dimensions
        if relative:
            self.list_points = [
                [
                    [
                        y - self.min_y,
                        x - self.min_x,
                    ]
                    for x, y in track_frame_points
                ]
                for track_frame_points in self.list_points
            ]
        else:
            self.list_points = [
                [[y, x] for x, y in track_frame_points]
                for track_frame_points in self.list_points
            ]



"""Run with binary_bridges_cnn_inference environment to generate semi images."""

import os

from cnn_framework.utils.tools import save_tiff
from cnn_framework.utils.readers.tiff_reader import TiffReader

from developers.binary_bridges_cnn.model_params import BinaryBridgesModelParams
from developers.binary_bridges_cnn.bridges_parser import BridgesParser
from developers.binary_bridges_cnn.semi_images_generation.micro_tubules_augmentations_advanced import (
    MicroTubulesAugmentationAdvanced,
)
from developers.binary_bridges_cnn.semi_images_generation.mt_cut_detection_factory_advanced import (
    MtCutDetectionFactoryAdvanced,
)


def main(params, save_dir, debug_plot, circle_radius=11, diff_radius=4.54):

    factory = MtCutDetectionFactoryAdvanced()
    file_paths = os.listdir(params.data_dir)

    # Create the save directory if it does not exist
    os.makedirs(save_dir, exist_ok=True)

    for file_path in file_paths:
        # Read image
        file_path = os.path.join(params.data_dir, file_path)
        image = TiffReader(file_path).image.squeeze()  # CYX

        # Get the bridge type
        bridge_type = file_path.split("_")[-1].split(".")[0][-1]
        # string to int
        bridge_type = int(bridge_type)
        class_mode = bridge_type % 5

        # Create the coordinates list of the circle around the mid body spot
        # NB: first circle in list is the middle circle
        list_radius = [
            circle_radius,
            circle_radius + 0.25 * diff_radius,
            circle_radius - 0.25 * diff_radius,
            circle_radius + 0.5 * diff_radius,
            circle_radius - 0.5 * diff_radius,
            circle_radius + 0.75 * diff_radius,
            circle_radius - 0.75 * diff_radius,
            circle_radius + diff_radius,
            circle_radius - diff_radius,
        ]

        # Apply image pre-processing
        sir_tubulin_image = image[0, ...].squeeze()  # YX

        # Get useful data for each circle
        all_intensities = []
        for radius in list_radius:
            intensities = factory.get_circle_data(radius, sir_tubulin_image)
            all_intensities.append(intensities)

        # Compute 2 best peaks on average circle
        average_circle_peaks = factory.get_average_circle_peaks(
            all_intensities, debug_plot=debug_plot
        )

        # Save fake one cut image
        if save_dir is not None:
            file_name = os.path.basename(file_path).split("_c")[0]

            # Keep number of peaks depending on class mode
            if class_mode in (1, 3):
                average_circle_peaks = average_circle_peaks[:1]
            if class_mode in (2, 4):
                average_circle_peaks = []

            micro_tubules_augmentation = MicroTubulesAugmentationAdvanced(
                average_circle_peaks
            )
            augmentations = micro_tubules_augmentation.generate_augmentations(
                image
            )
            for title, title_values in augmentations.items():
                category = title_values["category"]
                save_path = os.path.join(
                    save_dir, f"{file_name}_{title}_c{category}.tiff"
                )
                save_tiff(
                    title_values["image"], save_path, original_order="YXC"
                )


if __name__ == "__main__":
    parser = BridgesParser()
    args = parser.arguments_parser.parse_args()

    parameters = BinaryBridgesModelParams()
    parameters.update(args)

    SAVE_DIR = parameters.data_dir + "_semi_images"
    DEBUG_PLOT = False

    main(parameters, save_dir=SAVE_DIR, debug_plot=DEBUG_PLOT)



import json
import os
import numpy as np

from ..utils.metaphase_sequence import MetaphaseSequence
from ..utils.cell_track import CellTrack
from ..utils.tools import perform_cnn_inference
from ..utils.mitosis_track import MitosisTrack
from ..utils.cell_spot import CellSpot
from ..utils.hidden_markov_models import HiddenMarkovModel
from ..utils.mitosis_track_generation.metaphase_cnn_model_params import (
    MetaphaseCnnModelParams,
)
from ..utils.mitosis_track_generation.metaphase_cnn_data_set import (
    MetaphaseCnnDataSet,
)
from ..utils.mitosis_track_generation.metaphase_cnn_model import (
    MetaphaseCnnModel,
)
from ..constants.tracking import TIME_RESOLUTION


def get_track_from_id(tracks: list[CellTrack], track_id: int) -> CellTrack:
    """Used to get track from its id.

    Parameters
    ----------
    tracks : list[CellTrack]
        List of all tracks.
    track_id : int
        Track id.

    Returns
    -------
    CellTrack
        Track corresponding to track_id.
    """
    for track in tracks:
        if track.track_id == track_id:
            return track
    raise ValueError(f"Track {track_id} not found")


class MitosisTrackGenerationFactory:
    """Class to merge cell tracks into mitosis tracks.

    Parameters
    ----------
    minimum_metaphase_interval : int
        Minimum frames distance between two metaphases.
    max_spot_distance_for_split : int
        Maximum distance between two spots to consider them.
    """

    def __init__(
        self,
        minimum_metaphase_interval=100 / TIME_RESOLUTION,
        max_spot_distance_for_split=20,
    ) -> None:
        self.minimum_metaphase_interval = minimum_metaphase_interval
        self.max_spot_distance_for_split = max_spot_distance_for_split

    def get_tracks_to_merge(
        self, raw_tracks: list[CellTrack]
    ) -> list[MitosisTrack]:
        """Plug tracks occurring at frame>0 to closest metaphase.

        Parameters
        ----------
        raw_tracks : list[CellTrack]
            List of all tracks.

        Returns
        -------
        list[MitosisTrack]
            List of mitosis tracks.
        """
        ordered_tracks = sorted(raw_tracks, key=lambda x: x.start)
        mitosis_tracks: list[MitosisTrack] = []

        # Loop through all tracks beginning at frame > 0 and try to plug them to the previous metaphase
        for track in reversed(ordered_tracks):
            # Break when reaching tracking starting at first frame, as they are ordered
            track_first_frame = min(track.spots.keys())
            if track_first_frame == 0:
                break

            # Get all spots at same frame
            contemporary_spots = [
                raw_track.spots[track_first_frame]
                for raw_track in raw_tracks
                if track_first_frame in raw_track.spots
                and raw_track.track_id != track.track_id
            ]

            # Keep only stuck spots
            first_spot = track.spots[track_first_frame]
            stuck_spots: list[CellSpot] = list(
                filter(
                    lambda x: x.is_stuck_to(
                        first_spot, self.max_spot_distance_for_split
                    ),
                    contemporary_spots,
                )
            )

            # Keep only spots with metaphase sequence close to track first frame
            metaphase_spots = list(
                filter(
                    lambda x: get_track_from_id(
                        raw_tracks, x.track_id
                    ).has_close_metaphase(x, track_first_frame),
                    stuck_spots,
                )
            )

            # If no candidate has been found, ignore track
            if len(metaphase_spots) == 0:
                continue

            # Order remaining spots by overlap
            selected_spot = sorted(
                metaphase_spots,
                key=lambda x: get_track_from_id(
                    raw_tracks, x.track_id
                ).compute_metaphase_iou(track),
            )[-1]

            metaphase_sequence: MetaphaseSequence = (
                selected_spot.corresponding_metaphase_sequence
            )

            # Check if it should be merged to existing split (division into 3 cells)
            triple_division = False
            for mitosis_track in mitosis_tracks:
                if metaphase_sequence.is_same(
                    mitosis_track.metaphase_sequence
                ):
                    mitosis_track.add_daughter_track(track.track_id)
                    triple_division = True
                    break

            # If not, create new split
            if not triple_division:
                mitosis_tracks.append(
                    MitosisTrack(
                        track.track_id,
                        metaphase_sequence,
                    )
                )

        # Return dictionaries of tracks to merge
        return mitosis_tracks

    def _correct_sequence(self, orig_sequence: list[int]) -> list[int]:
        """Correct sequence of states to fill the gap between two metaphase (1) subsequences
        separated by less than minimum_interval frames.

        Parameters
        ----------
        orig_seq : list[int]
            Predicted classes.

        Returns
        -------
        list[int]
            Corrected classes.
        """
        corrected_sequence = np.copy(orig_sequence)
        # Get indexes of 1
        metaphase_index = [
            i for i, x in enumerate(corrected_sequence) if x == 1
        ]
        for idx in range(1, len(metaphase_index)):
            if (
                metaphase_index[idx] - metaphase_index[idx - 1] != 1
                and metaphase_index[idx] - metaphase_index[idx - 1]
                < self.minimum_metaphase_interval
            ):
                corrected_sequence[
                    list(range(metaphase_index[idx - 1], metaphase_index[idx]))
                ] = 1
        return corrected_sequence

    def pre_process_spots(
        self,
        cell_tracks: list[CellTrack],
        raw_spots: list[CellSpot],
        raw_video: np.ndarray,
        metaphase_model_path: str,
        hmm_metaphase_parameters_file: str,
        predictions_file: str,
        video_name: str,
        only_predictions_update: bool,
    ) -> None:
        """Sort spots in track and predict metaphase.

        Parameters
        ----------
        cell_tracks : list[CellTrack]
            List of cell tracks.
        raw_spots : list[CellSpot]
            List of all spots.
        raw_video : np.ndarray
            Video. TYXC.
        metaphase_model_path : str
            CNN model path.
        hmm_metaphase_parameters_file : str
            HMM parameters file.
        predictions_file : str
            Predictions file to update if provided.
        video_name : str
            Video name.
        only_predictions_update : bool
            If True, only update predictions, do not apply HMM.
        """

        nuclei_crops, numbers_crops = [], []
        # Get list of possible metaphase spots
        for track in cell_tracks:
            # Get current track spots data & images
            current_nuclei_crops = track.get_spots_data(raw_spots, raw_video)
            # Merge current_nuclei_crops with nuclei_crops
            nuclei_crops = nuclei_crops + current_nuclei_crops  # CYX
            numbers_crops.append(len(current_nuclei_crops))

        # Apply CNN model to get metaphase spots, once for all
        predictions = self._predict_metaphase_spots(
            metaphase_model_path, nuclei_crops
        )

        # Load HMM parameters and create model
        if not only_predictions_update:
            if not os.path.exists(hmm_metaphase_parameters_file):
                raise FileNotFoundError(
                    f"File {hmm_metaphase_parameters_file} not found"
                )
            hmm_parameters = np.load(hmm_metaphase_parameters_file)
            hmm_model = HiddenMarkovModel(
                hmm_parameters["A"], hmm_parameters["B"], hmm_parameters["pi"]
            )

        # Get list of possible metaphase spots
        for track, number_crops in zip(cell_tracks, numbers_crops):
            track_predictions = predictions[:number_crops]
            predictions = predictions[number_crops:]

            # Hidden Markov Model to smooth predictions
            # If we just want to get raw CNN predictions, we don't want to correct the predictions
            if not only_predictions_update:
                track_predictions, _ = hmm_model.viterbi_inference(
                    track_predictions
                )
                track_predictions = self._correct_sequence(track_predictions)

            # Save prediction for each spot
            track.update_metaphase_spots(track_predictions)

        self._update_predictions_file(
            cell_tracks, predictions_file, video_name
        )

    @staticmethod
    def _predict_metaphase_spots(
        metaphase_model_path: str, nuclei_crops: list[np.array]
    ) -> list[int]:
        """Run CNN model to predict metaphase spots.

        Parameters
        ----------
        metaphase_model : str
            CNN model path.
        nuclei_crops :  list[np.array]
            list[CYX]

        Returns
        -------
        list[int]
            Predicted classes.
        """

        predictions = perform_cnn_inference(
            model_path=metaphase_model_path,
            images=nuclei_crops,
            cnn_model_params=MetaphaseCnnModelParams,
            cnn_data_set=MetaphaseCnnDataSet,
            cnn_classifier=MetaphaseCnnModel,
        )
        return predictions

    @staticmethod
    def _update_predictions_file(
        tracks: list[CellTrack], predictions_file: str, video_name: str
    ) -> None:
        """Update predictions file with new predictions.
        Used only to train HMM model.

        Parameters
        ----------
        tracks: [CellTrack]
            List of cell tracks.
        predictions_file: str
            Predictions file to update.
        video_name: str
            Video name.
        """
        if predictions_file is None:
            return

        # Read data from json prediction if exists
        if os.path.exists(predictions_file):
            with open(predictions_file) as json_file:
                predictions_data = json.load(json_file)
        else:
            predictions_data = {}

        # Retrieve predictions
        predictions = {
            int(track.track_id): [
                int(spot.predicted_phase) for spot in track.spots.values()
            ]
            for track in tracks
        }
        predictions_data[video_name] = predictions
        # Save predictions data
        with open(predictions_file, "w") as json_file:
            json.dump(predictions_data, json_file)



import albumentations as A
import numpy as np

from cnn_framework.utils.data_sets.abstract_data_set import AbstractDataSet
from cnn_framework.utils.readers.images_reader import ImagesReader
from cnn_framework.utils.data_sets.dataset_output import DatasetOutput
from cnn_framework.utils.enum import ProjectMethods
from cnn_framework.utils.tools import read_categories_probability_from_name
from cnn_framework.utils.tools import to_one_hot
from cnn_framework.utils.readers.utils.projection import Projection
from cnn_framework.utils.file_name_encoder import FileNameEncoder

from developers.binary_bridges_cnn.tools import get_category_from_name


class BridgesCnnDataSet(AbstractDataSet):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Data sources
        self.input_data_source = ImagesReader(
            [self.data_manager.get_microscopy_image_path],
            [
                [
                    Projection(
                        method=ProjectMethods.Channel,
                        channels=self.params.c_indexes,
                        axis=1,
                    )
                ]
            ],
        )  # SiRTubulin, MKLP1, PC

    def set_transforms(self):
        height, width = self.params.input_dimensions.to_tuple(False)
        if self.is_train:
            self.transforms = A.Compose(
                [
                    A.Normalize(
                        self.mean_std["mean"],
                        std=self.mean_std["std"],
                        max_pixel_value=1,
                    ),
                    A.PadIfNeeded(
                        min_height=height,
                        min_width=width,
                        border_mode=0,
                        value=0,
                        p=1,
                    ),
                    A.CenterCrop(height=height, width=width, p=1),
                    A.Rotate(border_mode=0),
                    A.HorizontalFlip(),
                    A.VerticalFlip(),
                ]
            )
        else:
            self.transforms = A.Compose(
                [
                    A.Normalize(
                        self.mean_std["mean"],
                        std=self.mean_std["std"],
                        max_pixel_value=1,
                    ),
                    A.PadIfNeeded(
                        min_height=height,
                        min_width=width,
                        border_mode=0,
                        value=0,
                        p=1,
                    ),
                    A.CenterCrop(height=height, width=width, p=1),
                ]
            )

    def read_output(self, filename, one_hot=False):
        """
        Used for classification models.
        Read category and/or probabilities from file name
        """
        if self.params.mode == "microtubules":
            assert self.params.nb_classes == 3
        elif self.params.mode == "membrane":
            assert self.params.nb_classes == 2
        else:
            raise ValueError(f"Unknown mode: {self.params.mode}")

        # Read category from name
        categories_and_probabilities = filename.split(".")[0].split("_c")[1:]
        initial_category = int(categories_and_probabilities[0])

        # Match uncertain classes to basic ones
        category = get_category_from_name(
            initial_category, mode=self.params.mode
        )

        # Are there probabilities for classes?
        if len(categories_and_probabilities) > 1 and not one_hot:
            _, probabilities = read_categories_probability_from_name(filename)
        else:
            probabilities = to_one_hot(category, self.params.nb_classes)

        # Length of probabilities has to be equal to number of classes
        assert self.params.nb_classes == len(probabilities)

        # Category has to be the index of the highest probability
        assert category == np.argmax(probabilities)

        return np.asarray(probabilities)

    def generate_images(self, filename):
        # Output
        probabilities = self.read_output(filename)

        encoder = FileNameEncoder(
            self.params.names_train,
            self.params.names_val,
            self.params.names_test,
        )
        return DatasetOutput(
            input=self.input_data_source.get_image(filename),
            target_array=probabilities,
            additional=self.additional_data_source.get_image(filename),
            encoded_file_name=encoder.encode(filename),
        )



from typing import Optional

from cut_detector.utils.mt_cut_detection.micro_tubules_augmentation import (
    MicroTubulesAugmentation,
)

from developers.binary_bridges_cnn.semi_images_generation.peak import Peak


class MicroTubulesAugmentationAdvanced(MicroTubulesAugmentation):
    """
    Manage the augmentation of images for microtubules detection.
    """

    def __init__(self, peaks: Optional[list[Peak]] = None):
        super().__init__()
        peak_augmentations = []
        if peaks is None:
            peaks = []
        for peak in peaks:
            peak_augmentations.append(peak.enabled_augmentation())
        self.augmentations = self.merge_augmentations(peak_augmentations)

    @classmethod
    def merge_augmentations(
        cls, peak_augmentations: Optional[list[dict[str, int]]] = None
    ) -> dict[str, int]:
        """
        Merge values given to augmentation categories by different peaks.
        """
        augmentations = {}
        for category in [
            "top",
            "bottom",
            "left",
            "right",
            "top_left",
            "bottom_right",
            "top_right",
            "bottom_left",
        ]:
            value = 0  # by default, no MT is seen
            if peak_augmentations is not None:
                for (
                    peak_augmentation
                ) in peak_augmentations:  # iterate over peaks
                    if (
                        peak_augmentation[category] == -1
                    ):  # if one peak forbids the augmentation
                        value = None
                        break
                    if (
                        peak_augmentation[category] == 1
                    ):  # if at least one peak found
                        value = 1
            if value is not None:
                augmentations[category] = value
        return augmentations
